"""
# [How can I attach documentation to members of a python enum?](https://stackoverflow.com/questions/50473951/how-can-i-attach-documentation-to-members-of-a-python-enum/50473952#50473952)
"""

from enum import Enum


class DocEnum(Enum):
    def __new__(cls, value, doc=None):  # type: ignore
        self = object.__new__(cls)
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self

    def __call__(self, *args, **kwargs):  # type: ignore
        return self.value(*args, **kwargs)



import binascii
from typing import Callable, Final

import imagehash
import numpy as np
from imagehash import ImageHash
from PIL import Image

FRAME_SIZE: Final[int] = 32


def imghash_to_bytes(imghash: ImageHash) -> bytes:
    """
    >>> import numpy as np
    >>> img_hash = ImageHash(np.array([\
        np.array([ True,  True, False,  True, False,  True, False,  True]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False]), \
        np.array([False, False, False, False, False, False, False, False])]))
    >>> str(img_hash)
    'd500000000000000'
    >>> imghash_to_bytes(img_hash)
    b'\\xd5\\x00\\x00\\x00\\x00\\x00\\x00\\x00'
    """
    bin_array_to_hex = str(imghash)
    return binascii.a2b_hex(bin_array_to_hex)


def bytes_to_imghash(raw_bytes: bytes) -> ImageHash:
    """

    Args:
        raw_bytes (bytes): raw bytes to hash

    Returns:
        ImageHash: Image hash representation of binaries input

    Examples:
        >>> import numpy as np
        >>> img_hash = ImageHash(np.array([\
            np.array([ True,  True, False,  True, False,  True, False,  True]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False]), \
            np.array([False, False, False, False, False, False, False, False])]))
        >>> imghash_reconstructed = bytes_to_imghash(b'\\xd5\\x00\\x00\\x00\\x00\\x00\\x00\\x00')
        >>> str(imghash_reconstructed)
        'd500000000000000'
    """
    return imagehash.hex_to_hash(binascii.b2a_hex(raw_bytes).decode())


def rawframe_to_imghash(
    raw_frame: bytes,
    frame_width: int = FRAME_SIZE,
    frame_height: int = FRAME_SIZE,
    fn_imagehash: Callable[[Image.Image], ImageHash] = imagehash.phash,
) -> imagehash.ImageHash:
    """Apply a Perceptual Hash computation on raw image frame.

    >>> rawframe_to_imghash(bytearray((128,)*(FRAME_SIZE**2)))
    array([[ True, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False, False]])
    """
    return fn_imagehash(
        Image.fromarray(
            # https://docs.scipy.org/doc/numpy-1.14.0/reference/generated/numpy.frombuffer.html
            np.frombuffer(
                raw_frame, dtype=np.uint8, count=frame_width * frame_height
            ).reshape(frame_width, frame_height)
        )
    )



import subprocess
import threading
import time
from typing import BinaryIO

from imageio_ffmpeg import get_ffmpeg_exe
from imageio_ffmpeg._parsing import LogCatcher, parse_ffmpeg_header
from imageio_ffmpeg._utils import _popen_kwargs, logger


def read_frames_from_binary_stream(
    bin_io_stream: BinaryIO,
    pix_fmt="rgb24",
    bpp=None,
    input_params=None,
    output_params=None,
    bits_per_pixel=None,
    chunk_size_for_input_stream_reading: int = 8_192,
):
    """
    Create a generator to iterate over the frames in a video file.

    It first yields a small metadata dictionary that contains:

    * ffmpeg_version: the ffmpeg version in use (as a string).
    * codec: a hint about the codec used to encode the video, e.g. "h264".
    * source_size: the width and height of the encoded video frames.
    * size: the width and height of the frames that will be produced.
    * fps: the frames per second. Can be zero if it could not be detected.
    * duration: duration in seconds. Can be zero if it could not be detected.

    After that, it yields frames until the end of the video is reached. Each
    frame is a bytes object.

    This function makes no assumptions about the number of frames in
    the data. For one because this is hard to predict exactly, but also
    because it may depend on the provided output_params. If you want
    to know the number of frames in a video file, use count_frames_and_secs().
    It is also possible to estimate the number of frames from the fps and
    duration, but note that even if both numbers are present, the resulting
    value is not always correct.

    Example:

        gen = read_frames(path)
        meta = gen.__next__()
        for frame in gen:
            print(len(frame))

    Parameters:
        bin_io_stream (BinaryIO): .
        pix_fmt (str): the pixel format of the frames to be read.
            The default is "rgb24" (frames are uint8 RGB images).
        input_params (list): Additional ffmpeg input command line parameters.
        output_params (list): Additional ffmpeg output command line parameters.
        bits_per_pixel (int): The number of bits per pixel in the output frames.
            This depends on the given pix_fmt. Default is 24 (RGB)
        bpp (int): DEPRECATED, USE bits_per_pixel INSTEAD. The number of bytes per pixel in the output frames.
            This depends on the given pix_fmt. Some pixel formats like yuv420p have 12 bits per pixel
            and cannot be set in bytes as integer. For this reason the bpp argument is deprecated.
        chunk_size_for_input_stream_reading (int): size (in bits) used for chunk reading from input stream
    """

    # ----- Input args

    # if isinstance(path, pathlib.PurePath):
    #     path = str(path)
    # if not isinstance(path, str):
    #     raise TypeError("Video path must be a string or pathlib.Path.")
    # # Note: Don't check whether it exists. The source could be e.g. a camera.

    pix_fmt = pix_fmt or "rgb24"
    bpp = bpp or 3
    bits_per_pixel = bits_per_pixel or bpp * 8
    input_params = input_params or []
    output_params = output_params or []

    assert isinstance(pix_fmt, str), "pix_fmt must be a string"
    assert isinstance(bits_per_pixel, int), "bpp and bits_per_pixel must be an int"
    assert isinstance(input_params, list), "input_params must be a list"
    assert isinstance(output_params, list), "output_params must be a list"

    # ----- Prepare

    pre_output_params = ["-pix_fmt", pix_fmt, "-vcodec", "rawvideo", "-f", "image2pipe"]

    cmd = [get_ffmpeg_exe()]
    # [Pipe input in to ffmpeg stdin](https://stackoverflow.com/a/45902691)
    # [ffmpeg-protocols.html#pipe](https://ffmpeg.org/ffmpeg-protocols.html#pipe)
    cmd += input_params + ["-i", "pipe:0"]
    cmd += pre_output_params + output_params + ["-"]

    process = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **_popen_kwargs(prevent_sigint=True),
    )

    # Thread to inject input stream by chunks
    # [How do I write to a Python subprocess' stdin?](https://stackoverflow.com/a/76708895)
    # [Popen.stdin](https://docs.python.org/3/library/subprocess.html#subprocess.Popen.stdin)
    # [Popen's communicate vs stdin.write: Why warning? What instead?](https://www.reddit.com/r/learnpython/comments/17c7h7u/popens_communicate_vs_stdinwrite_why_warning_what/)
    # [cpython/Lib/subprocess.py: def communicate(...)](https://github.com/python/cpython/blob/main/Lib/subprocess.py#L1174)
    # [Send input from one threaded subprocess to another](https://stackoverflow.com/questions/41287291/send-input-from-one-threaded-subprocess-to-another)
    def write_to_input_stream(_process, _bin_io_stream: BinaryIO):
        logger.info(
            f"starting to write to input stream ({chunk_size_for_input_stream_reading=} octets)"
        )
        # [Read file in chunks - RAM-usage, reading strings from binary files](https://stackoverflow.com/a/63563264)
        try:
            while chunk := _bin_io_stream.read(chunk_size_for_input_stream_reading):
                _process.stdin.write(chunk)
                _process.stdin.flush()
        # to prevent error occurred in doctest (at the end)
        except BrokenPipeError:
            pass
        finally:
            _process.stdin.close()
            pass

    thread_write_to_input_stream = threading.Thread(
        target=write_to_input_stream, args=(process, bin_io_stream)
    )
    # Start a thread for injecting binary data to input
    thread_write_to_input_stream.daemon = True
    thread_write_to_input_stream.start()

    log_catcher = LogCatcher(process.stderr)

    # Init policy by which to terminate ffmpeg. May be set to "kill" later.
    stop_policy = "timeout"  # not wait; ffmpeg should be able to quit quickly

    # Enter try block directly after opening the process.
    # We terminate ffmpeg in the final clause.
    # Generators are automatically closed when they get deleted,
    # so the finally block is guaranteed to run.
    try:
        # ----- Load meta data

        # Wait for the log catcher to get the meta information
        etime = time.time() + 10.0
        while log_catcher.is_alive() and not log_catcher.header and time.time() < etime:
            time.sleep(0.01)

        # Check whether we have the information
        if not log_catcher.header:
            err2 = log_catcher.get_text(0.2)
            fmt = "Could not load meta information\n=== stderr ===\n{}"
            raise IOError(fmt.format(err2))
        # TODO: maybe an another log error type for pipe input ... need to look at it !
        # elif "No such file or directory" in log_catcher.header:
        #     raise IOError("{} not found! Wrong path?".format(path))

        meta = parse_ffmpeg_header(log_catcher.header)
        yield meta

        # ----- Read frames

        width, height = meta["size"]
        framesize_bits = width * height * bits_per_pixel
        framesize_bytes = framesize_bits / 8
        assert (
            framesize_bytes.is_integer()
        ), "incorrect bits_per_pixel, framesize in bytes must be an int"
        framesize_bytes = int(framesize_bytes)
        framenr = 0

        while True:
            framenr += 1
            try:
                bb = bytes()
                while len(bb) < framesize_bytes:
                    extra_bytes = process.stdout.read(framesize_bytes - len(bb))
                    if not extra_bytes:
                        if len(bb) == 0:
                            return
                        else:
                            raise RuntimeError(
                                "End of file reached before full frame could be read."
                            )
                    bb += extra_bytes
                yield bb
            except Exception as err:
                err1 = str(err)
                err2 = log_catcher.get_text(0.4)
                fmt = "Could not read frame {}:\n{}\n=== stderr ===\n{}"
                raise RuntimeError(fmt.format(framenr, err1, err2))

    except GeneratorExit:
        # Note that GeneratorExit does not inherit from Exception but BaseException
        pass

    except Exception:
        # Normal exceptions fall through
        raise

    except BaseException:
        # Detect KeyboardInterrupt / SystemExit: don't wait for ffmpeg to quit
        stop_policy = "kill"
        raise

    finally:
        # waiting the end of the (Producer) thread
        try:
            thread_write_to_input_stream.join(timeout=1)
        except TimeoutError:
            logger.warning(
                "Consumer (FFMPEG stdout) stop before Producer (FFMPEG stdin)."
            )

        # Stop the LogCatcher thread, which reads from stderr.
        log_catcher.stop_me()

        # Make sure that ffmpeg is terminated.
        if process.poll() is None:
            # Ask ffmpeg to quit
            try:
                # I read somewhere that modern ffmpeg on Linux prefers a
                # "ctrl-c", but tests so far suggests sending q is more robust.
                # > p.send_signal(signal.SIGINT)
                # Sending q via communicate works, but can hang (see #17)
                # > p.communicate(b"q")
                # So let's do similar to what communicate does, but without
                # reading stdout (which may block). It looks like only closing
                # stdout is enough (tried Windows+Linux), but let's play safe.
                # Found that writing to stdin can cause "Invalid argument" on
                # Windows # and "Broken Pipe" on Unix.
                # p.stdin.write(b"q")  # commented out in v0.4.1
                process.stdout.close()
                # process.stdin.close() -> not here, the thread writing into input stream closes it
                # p.stderr.close() -> not here, the log_catcher closes it
            except Exception as err:  # pragma: no cover
                logger.warning("Error while attempting stop ffmpeg (r): " + str(err))

            if stop_policy == "timeout":
                # Wait until timeout, produce a warning and kill if it still exists
                try:
                    etime = time.time() + 1.5
                    while time.time() < etime and process.poll() is None:
                        time.sleep(0.01)
                finally:
                    if process.poll() is None:  # pragma: no cover
                        logger.warning("We had to kill ffmpeg to stop it.")
                        process.kill()

            else:  # stop_policy == "kill"
                # Just kill it
                process.kill()


# TODO: need to mutualise code (too much duplications)
def read_frames_from_url(
    url,
    pix_fmt="rgb24",
    bpp=None,
    input_params=None,
    output_params=None,
    bits_per_pixel=None,
):
    """
    Create a generator to iterate over the frames in a video file.

    It first yields a small metadata dictionary that contains:

    * ffmpeg_version: the ffmpeg version in use (as a string).
    * codec: a hint about the codec used to encode the video, e.g. "h264".
    * source_size: the width and height of the encoded video frames.
    * size: the width and height of the frames that will be produced.
    * fps: the frames per second. Can be zero if it could not be detected.
    * duration: duration in seconds. Can be zero if it could not be detected.

    After that, it yields frames until the end of the video is reached. Each
    frame is a bytes object.

    This function makes no assumptions about the number of frames in
    the data. For one because this is hard to predict exactly, but also
    because it may depend on the provided output_params. If you want
    to know the number of frames in a video file, use count_frames_and_secs().
    It is also possible to estimate the number of frames from the fps and
    duration, but note that even if both numbers are present, the resulting
    value is not always correct.

    Example:

        gen = read_frames(path)
        meta = gen.__next__()
        for frame in gen:
            print(len(frame))

    Parameters:
        url (str): url pointing to streaming video (for example from YouTube).
        pix_fmt (str): the pixel format of the frames to be read.
            The default is "rgb24" (frames are uint8 RGB images).
        input_params (list): Additional ffmpeg input command line parameters.
        output_params (list): Additional ffmpeg output command line parameters.
        bits_per_pixel (int): The number of bits per pixel in the output frames.
            This depends on the given pix_fmt. Default is 24 (RGB)
        bpp (int): DEPRECATED, USE bits_per_pixel INSTEAD. The number of bytes per pixel in the output frames.
            This depends on the given pix_fmt. Some pixel formats like yuv420p have 12 bits per pixel
            and cannot be set in bytes as integer. For this reason the bpp argument is deprecated.
    """

    # ----- Input args
    pix_fmt = pix_fmt or "rgb24"
    bpp = bpp or 3
    bits_per_pixel = bits_per_pixel or bpp * 8
    input_params = input_params or []
    output_params = output_params or []

    assert isinstance(pix_fmt, str), "pix_fmt must be a string"
    assert isinstance(bits_per_pixel, int), "bpp and bits_per_pixel must be an int"
    assert isinstance(input_params, list), "input_params must be a list"
    assert isinstance(output_params, list), "output_params must be a list"

    # ----- Prepare

    pre_output_params = ["-pix_fmt", pix_fmt, "-vcodec", "rawvideo", "-f", "image2pipe"]

    cmd = [get_ffmpeg_exe()]
    cmd += input_params + ["-i", url]
    cmd += pre_output_params + output_params + ["-"]

    process = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **_popen_kwargs(prevent_sigint=True),
    )

    log_catcher = LogCatcher(process.stderr)

    # Init policy by which to terminate ffmpeg. May be set to "kill" later.
    stop_policy = "timeout"  # not wait; ffmpeg should be able to quit quickly

    # Enter try block directly after opening the process.
    # We terminate ffmpeg in the final clause.
    # Generators are automatically closed when they get deleted,
    # so the finally block is guaranteed to run.
    try:
        # ----- Load meta data

        # Wait for the log catcher to get the meta information
        etime = time.time() + 10.0
        while log_catcher.is_alive() and not log_catcher.header and time.time() < etime:
            time.sleep(0.01)

        # Check whether we have the information
        if not log_catcher.header:
            err2 = log_catcher.get_text(0.2)
            fmt = "Could not load meta information\n=== stderr ===\n{}"
            raise IOError(fmt.format(err2))
        # elif "No such file or directory" in log_catcher.header:
        #     raise IOError("{} not found! Wrong path?".format(path))

        meta = parse_ffmpeg_header(log_catcher.header)
        yield meta

        # ----- Read frames

        width, height = meta["size"]
        framesize_bits = width * height * bits_per_pixel
        framesize_bytes = framesize_bits / 8
        assert (
            framesize_bytes.is_integer()
        ), "incorrect bits_per_pixel, framesize in bytes must be an int"
        framesize_bytes = int(framesize_bytes)
        framenr = 0

        while True:
            framenr += 1
            try:
                bb = bytes()
                while len(bb) < framesize_bytes:
                    extra_bytes = process.stdout.read(framesize_bytes - len(bb))
                    if not extra_bytes:
                        if len(bb) == 0:
                            return
                        else:
                            raise RuntimeError(
                                "End of file reached before full frame could be read."
                            )
                    bb += extra_bytes
                yield bb
            except Exception as err:
                err1 = str(err)
                err2 = log_catcher.get_text(0.4)
                fmt = "Could not read frame {}:\n{}\n=== stderr ===\n{}"
                raise RuntimeError(fmt.format(framenr, err1, err2))

    except GeneratorExit:
        # Note that GeneratorExit does not inherit from Exception but BaseException
        pass

    except Exception:
        # Normal exceptions fall through
        raise

    except BaseException:
        # Detect KeyboardInterrupt / SystemExit: don't wait for ffmpeg to quit
        stop_policy = "kill"
        raise

    finally:
        # Stop the LogCatcher thread, which reads from stderr.
        log_catcher.stop_me()

        # Make sure that ffmpeg is terminated.
        if process.poll() is None:
            # Ask ffmpeg to quit
            try:
                # I read somewhere that modern ffmpeg on Linux prefers a
                # "ctrl-c", but tests so far suggests sending q is more robust.
                # > p.send_signal(signal.SIGINT)
                # Sending q via communicate works, but can hang (see #17)
                # > p.communicate(b"q")
                # So let's do similar to what communicate does, but without
                # reading stdout (which may block). It looks like only closing
                # stdout is enough (tried Windows+Linux), but let's play safe.
                # Found that writing to stdin can cause "Invalid argument" on
                # Windows # and "Broken Pipe" on Unix.
                # p.stdin.write(b"q")  # commented out in v0.4.1
                process.stdout.close()
                process.stdin.close()
                # p.stderr.close() -> not here, the log_catcher closes it
            except Exception as err:  # pragma: no cover
                logger.warning("Error while attempting stop ffmpeg (r): " + str(err))

            if stop_policy == "timeout":
                # Wait until timeout, produce a warning and kill if it still exists
                try:
                    etime = time.time() + 1.5
                    while time.time() < etime and process.poll() is None:
                        time.sleep(0.01)
                finally:
                    if process.poll() is None:  # pragma: no cover
                        logger.warning("We had to kill ffmpeg to stop it.")
                        process.kill()

            else:  # stop_policy == "kill"
                # Just kill it
                process.kill()



import datetime
from io import BufferedReader
from pathlib import Path
from typing import Any, BinaryIO, Iterator, Tuple, Union

from imageio_ffmpeg import count_frames_and_secs, read_frames

from vhcalc.models import URL, MetaData
from vhcalc.tools.forked.imageio_ffmpeg_io import (
    read_frames_from_binary_stream,
    read_frames_from_url,
)
from vhcalc.tools.imghash import FRAME_SIZE


def build_reader_frames(
    media_input: Union[Path, Union[BufferedReader, BinaryIO], URL],
    nb_seconds_to_extract: float = 0,
    seek_to_middle: bool = False,
    ffmpeg_reduce_verbosity: bool = False,
) -> Tuple[Iterator[bytes], MetaData]:
    """

    Args:
        media_input:
        nb_seconds_to_extract:
        seek_to_middle:
        ffmpeg_reduce_verbosity:

    Returns:

    """
    ffmpeg_seek_input_cmd: list[str] = []
    ffmpeg_seek_output_cmd: list[str] = []
    s_media: str = ""

    # https://trac.ffmpeg.org/wiki/Seeking#Cuttingsmallsections
    if ffmpeg_reduce_verbosity:
        ffmpeg_seek_input_cmd += "-hide_banner -nostats -nostdin".split(" ")

    if isinstance(media_input, Path):
        s_media = str(media_input)
        # get media metadata from first yield of frames reader
        metadata_from_frames_reader: dict[str, Any] = next(read_frames(s_media))

        # extract a (frame's) chunk around/in middle of the media
        if seek_to_middle:
            # it's an approximation, this command seek around/close to the middle
            ffmpeg_seek_input_cmd += (
                "-ss",
                str(
                    datetime.timedelta(
                        seconds=metadata_from_frames_reader["duration"] // 2
                    )
                ),
            )

        if nb_seconds_to_extract:
            # express in number of frames to extract (more precise)
            ffmpeg_seek_output_cmd += (
                "-frames:v",
                str(round(nb_seconds_to_extract * metadata_from_frames_reader["fps"])),
            )

    # rescale output frame to 32x32
    video_filters = f"scale=width={FRAME_SIZE}:height={FRAME_SIZE}"

    if isinstance(media_input, Path):
        fn_read_frames = read_frames
    elif isinstance(media_input, BufferedReader):
        fn_read_frames = read_frames_from_binary_stream
    elif isinstance(media_input, URL):
        fn_read_frames = read_frames_from_url
    else:
        # TODO: handle this exception
        raise RuntimeError(f"Can't handle {type(media_input)=}")

    reader = fn_read_frames(
        media_input,
        pix_fmt="gray",
        input_params=[*ffmpeg_seek_input_cmd],
        output_params=[
            *ffmpeg_seek_output_cmd,
            *("-vf", video_filters),
        ],
        bits_per_pixel=8,
    )
    # get media metadata from first yield of reader
    metadata_from_frames_reader = next(reader)

    if isinstance(media_input, Path):
        nb_frames = count_frames_and_secs(s_media)[0]
    else:
        # FIXME: not accurate/exact => not working (very well)
        nb_frames = int(
            metadata_from_frames_reader["fps"] * metadata_from_frames_reader["duration"]
        )

    # build metadata dataclass
    metadata = MetaData(
        fps=metadata_from_frames_reader["fps"],
        duration=metadata_from_frames_reader["duration"],
        nb_frames=nb_frames,
    )

    return reader, metadata



from functools import partial
from io import BufferedReader, BytesIO
from pathlib import Path

# https://pypi.org/project/click-pathlib/
from tempfile import gettempdir
from typing import Iterable, Optional, Union

from imagehash import ImageHash
from rich import get_console

from vhcalc.models import URL, ImageHashingFunction
from vhcalc.services.reader_frames import build_reader_frames
from vhcalc.tools.chunk import chunks
from vhcalc.tools.imghash import bytes_to_imghash, imghash_to_bytes, rawframe_to_imghash
from vhcalc.tools.progress_bar import configure_progress_bar

console = get_console()


def b2b_stream_to_imghash(
    # FIXME: ugly need to refactor
    binary_stream: Union[BufferedReader, URL],
    chunk_size_in_frames: int = 15 * 25,
    fn_imagehash: ImageHashingFunction = ImageHashingFunction.PerceptualHashing,
) -> Iterable[bytes]:
    """
    Compute images hashes from file (media/video) binary content stream (send to ffmpeg)

    Args:
        binary_stream (BufferedReader): binary stream read from media file input
        chunk_size_in_frames (int): Chunk size in frames used for generating images hashes from media decompression.
        fn_imagehash (ImageHashingFunction): ImageHash function for transforming PIL.Image to ImageHash

    Yields:
        Iterable[bytes]: The next binary image hash from media input stream

    Example:
        >>> media_path = Path("tests/data/big_buck_bunny_trailer_480p.mkv")
        >>> next(b2b_stream_to_imghash(media_path.open("rb")))
        b'\xd5\xd5*\xd5*\xd4*\xd4'
    """
    # Read a video file
    it_reader_frame, _ = build_reader_frames(binary_stream)
    chunk_size = chunk_size_in_frames

    # configure chunk
    gen_imghashes = map(
        partial(rawframe_to_imghash, fn_imagehash=fn_imagehash), it_reader_frame
    )
    gen_chunk_imghashes = chunks(gen_imghashes, chunk_size)
    # for each chunk of frames
    for chunk_imghashes in gen_chunk_imghashes:
        for bin_imghash in map(imghash_to_bytes, chunk_imghashes):
            # and write (chunk of) images hashes result on export file
            yield bin_imghash


def a2b_imghash(
    binary_stream: BufferedReader,
    chunk_size: int = 8 * 1024,
) -> Iterable[ImageHash]:
    """

    Args:
        binary_stream (BufferedReader): expected binary stream to read compatible with images hashes binary format.
        chunk_size: size (in bits) used for chunk reading from input stream

    Returns:

    Examples:
        >>> bin_imghashes = b'\\xd5\\xd5*\\xd5*\\xd4*\\xd4' * 8
        >>> bin_stream = BytesIO(bin_imghashes)
        >>> imghash_reconstructed = next(a2b_imghash(bin_stream))
        >>> str(imghash_reconstructed)
        'd5d52ad52ad42ad4'
    """
    while chunk_bin_imghash := binary_stream.read(chunk_size):
        stream_bin_imghash = BytesIO(chunk_bin_imghash)
        while bin_imghash := stream_bin_imghash.read(8):
            yield bytes_to_imghash(bin_imghash)


def export_imghash_from_media(
    input_media: Path,
    output_file: Optional[Path] = None,
    chunk_nb_seconds: int = 15,
    unlink_export_file: bool = True,
) -> Path:
    """
    Export images hashes from media (readable with ffmpeg)

    Args:
        input_media (Path): Path object targeting the input media.
        output_file (Optional[Path]): Path object for the output file. If not given, a temporary file is created.
        chunk_nb_seconds (int): Chunk size in seconds used for generating images hashes from media decompression.
        unlink_export_file (bool): Option for apply Path.unlink() on output file.

    Returns:
        pathlib.Path: Path for the output file that contain binary images hashes.

    """
    # Read a video file
    it_reader_frame, media_metadata = build_reader_frames(input_media)
    nb_frames_to_read = media_metadata.nb_frames
    chunk_size = int(media_metadata.fps * chunk_nb_seconds)

    console.print(f"{media_metadata}")
    console.print(f"Number of frames to read: {nb_frames_to_read}")
    console.print(f"Chunk size (nb frames): {chunk_size}")

    # manage export
    if not output_file:
        # https://bandit.readthedocs.io/en/latest/plugins/b108_hardcoded_tmp_directory.html
        output_file = (
            Path(gettempdir()) / f"{input_media.name}.{media_metadata.fps}fps.phash"
        )

    # remove/unlink export file if exist
    if unlink_export_file:
        output_file.unlink(missing_ok=True)

    console.print(f"output_file: {str(output_file)}")

    # configure progress bar
    progress_bar, pb_task_id = configure_progress_bar(
        input_media.name, nb_frames_to_read * 8, console
    )
    pb_advance = chunk_size * 8

    with progress_bar:
        gen_imghashes = map(rawframe_to_imghash, it_reader_frame)
        gen_chunk_imghashes = chunks(gen_imghashes, chunk_size)
        # for each chunk of frames
        for chunk_imghashes in gen_chunk_imghashes:
            # ... open/write and close export file in binary append mode
            with output_file.open("ab") as fo:
                # consume generator: chunk frame to image hash
                for frame_hash_binary in map(imghash_to_bytes, chunk_imghashes):
                    # and write (chunk of) images hashes result on export file
                    fo.write(frame_hash_binary)
            # update progress bar synchronize with chunk progression
            progress_bar.update(pb_task_id, advance=pb_advance, refresh=True)
    return output_file



import sys

# to prevent deprecated warning emit by pytest:
#   > DeprecationWarning: pkg_resources is deprecated as an API.
#   > See https://setuptools.pypa.io/en/latest/pkg_resources.html
# https://docs.python.org/3.9/library/importlib.metadata.html#distributions
from importlib.metadata import distribution

from imageio_ffmpeg import get_ffmpeg_version


def get_version_extended_informations() -> str:
    """

    Returns:
        str: String representation of the version extended information:
            - python version: python version interpreter
            - ffmpeg version: ffmpeg version binary (with some builds information like gcc version/platform)
            - for each "core" dependencies package: imagehash, numpy and Pillow
            => name and version package (using by the application)

    Example:
        The CLI option `version` get information from `get_version_extended_informations` function

        $ vhcalc --version
        vhcalc, version 0.4.0
        python version: 3.9.7 (default, Nov  7 2021, 10:52:45)
        [GCC 10.2.1 20210110]
        ffmpeg version 4.2.2-static
        ImageHash: 4.3.1
        numpy: 2.0.2
        pillow: 10.4.0
    """
    return "\n".join(
        (
            f"python version: {sys.version}",
            f"ffmpeg version {get_ffmpeg_version()}",
            "\n".join(
                f"{pkg_distro.name}: {pkg_distro.version}"  # type: ignore[attr-defined]
                for pkg_distro in map(distribution, ("imagehash", "numpy", "pillow"))
            ),
        )
    )



"""
"""


class URL(str):
    def __init__(self, url: str):
        if not (
            url.startswith("https://")
            or url.startswith("http://")
            or url.startswith("ftp://")
        ):
            raise AssertionError(url)



# -*- coding: utf-8 -*-
import pathlib
import sys
from importlib.metadata import version
from io import BufferedReader, BufferedWriter
from typing import Iterable, Optional, Union

import rich_click as click
from loguru import logger

try:
    from pymediainfo import MediaInfo

    MediaInfo._get_library()
except OSError:
    # FIXME: on macos platform - OSError: Failed to load library from libmediainfo.0.dylib, libmediainfo.dylib - dlopen(libmediainfo.0.dylib, 0x0006)
    logger.error("Can't get library from pymediainfo.MediaInfo !", exc_info=True)

    import platform

    # this error is handle only on macOS platform (i.e "Darwin")
    if platform.system() != "Darwin":
        raise RuntimeError("Unexpected error occurred !")

    from unittest.mock import Mock

    mock_MediaInfo = Mock()
    mock_MediaInfo.parse = Mock(return_value={})
    MediaInfo = mock_MediaInfo

import vhcalc.services as services
from vhcalc.models import URL, ImageHashingFunction
from vhcalc.tools.forked.click_default_group import DefaultGroup
from vhcalc.tools.forked.click_path import GlobPaths
from vhcalc.tools.version_extended_informations import get_version_extended_informations


@click.version_option(
    version=version("vhcalc"),
    prog_name="vhcalc",
    message=f"%(prog)s, version %(version)s\n{get_version_extended_informations()}",
)
@click.group(
    cls=DefaultGroup,
    default="imghash",
    default_if_no_args=True,
    invoke_without_command=True,
)
def cli() -> None:
    pass


@cli.command(
    short_help="Compute image hashes from and to binaries stream (by default: stdin/out)"
)
@click.argument(
    "input_stream",
    type=click.File("rb"),
    default=sys.stdin.buffer,
)
@click.argument(
    "output_stream",
    type=click.File("wb"),
    default=sys.stdout.buffer,
)
@click.option(
    "--image-hashing-method",
    type=click.Choice(ImageHashingFunction.names()),
    default="PerceptualHashing",
    show_default=True,
    # TODO: post validation and transform this option string to callable image hashing function
    # see: [Python Enum support for click.Choice #605](https://github.com/pallets/click/issues/605#issuecomment-901099036)
    help="The image hashing method to use.",
)
@click.option(
    "--decompress",
    # https://click.palletsprojects.com/en/8.1.x/options/#boolean-flags
    is_flag=True,
    type=bool,
    default=False,
    help="Decompress images hashes from binary streams.",
)
@click.option(
    "--from-url",
    type=URL,
    help="Allow to pass an URL for INPUT",
)
def imghash(
    input_stream: BufferedReader,
    output_stream: BufferedWriter,
    image_hashing_method: str,
    decompress: bool,
    from_url: Optional[URL],
) -> None:
    """Generate images hashes from INPUT binary stream and send it to OUTPUT stream.

    INPUT stream (default: stdin)
    OUTPUT stream (default: stdout)
    """
    if decompress:
        for imghash_binary in services.a2b_imghash(input_stream):
            output_stream.write(str(imghash_binary).encode())
        return

    # FIXME: ugly need to refactor
    b2a_imghash_input: Union[BufferedReader, URL]
    if from_url:
        b2a_imghash_input = from_url
    else:
        b2a_imghash_input = input_stream
    for frame_hash_binary in services.b2b_stream_to_imghash(
        b2a_imghash_input,
        fn_imagehash=ImageHashingFunction[image_hashing_method],
    ):
        output_stream.write(frame_hash_binary)


@cli.command(short_help="Generate media informations (with `mediainfo` tool).")
@click.argument(
    "filename",
    type=str,
)
def mediainfo(filename: str) -> None:
    """
    FILENAME path to the media file or file-like object which will be analyzed.
        A URL can also be used if libmediainfo was compiled with CURL support.
    """
    json_mediainfo: str = MediaInfo.parse(filename, output="JSON")
    click.echo(json_mediainfo)


@cli.command(
    short_help="extracting and exporting binary video hashes (fingerprints) from any video source"
)
@click.option(
    "--medias_pattern",
    "-r",
    required=True,
    type=GlobPaths(
        files_okay=True,
        dirs_okay=False,
        readable_only=True,
        at_least_one=True,
    ),
    help="Pattern to find medias",
)
@click.option(
    "--output-file",
    "-o",
    default=None,
    type=click.Path(writable=True, path_type=pathlib.Path),
    help="File where to write images hashes.",
)
@logger.catch
def export_imghash_from_media(
    medias_pattern: Iterable[pathlib.Path], output_file: Optional[pathlib.Path]
) -> None:
    """Click entrypoint for extracting and exporting binary video hashes (fingerprints) from any video source"""
    for media in medias_pattern:
        services.export_imghash_from_media(media, output_file)


if __name__ == "__main__":
    cli()



import imagehash

from vhcalc.tools.docenum import DocEnum


class ImageHashingFunction(DocEnum):
    """
    >>> assert all(name_fn_img_hash in ImageHashingFunction.__doc__ for name_fn_img_hash in {"Average Hash computation",
    ... "Perceptual Hash computation.", "Difference Hash computation. computes differences horizontally",
    ... "Wavelet Hash computation."})
    >>>
    >>> from pathlib import Path
    >>> from vhcalc.services.imghashes import b2b_stream_to_imghash
    >>>
    >>> media_path = Path("tests/data/big_buck_bunny_trailer_480p.mkv")
    >>> it_imghash = b2b_stream_to_imghash(media_path.open("rb"),
    ... fn_imagehash=ImageHashingFunction.AverageHashing)
    >>> print(next(it_imghash))
    b'\xfe\xfe\xfe\xfe\xfe\xfe\xfe\xfe'
    """

    AverageHashing = imagehash.average_hash, "Average Hash computation"
    PerceptualHashing = imagehash.phash, "Perceptual Hash computation."
    PerceptualHashing_Simple = imagehash.phash_simple, "Perceptual Hash computation."
    DifferenceHashing = (
        imagehash.dhash,
        "Difference Hash computation. computes differences horizontally",
    )
    WaveletHashing = imagehash.whash, "Wavelet Hash computation."

    # TODO: return multiple images hash, not handle yet.
    # CropResistantHashing = imagehash.crop_resistant_hash

    # [How to get all values from python enum class?](https://stackoverflow.com/questions/29503339/how-to-get-all-values-from-python-enum-class)
    @classmethod
    def names(cls) -> list[str]:
        return list(map(lambda c: c.name, cls))



from vhcalc.models.imghash_function import ImageHashingFunction
from vhcalc.models.metadata import MetaData
from vhcalc.models.url import URL

__all__ = [
    "MetaData",
    "URL",
    "ImageHashingFunction",
]
